package ca.mcgill.sel.aspectoptima;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.*;
import ca.mcgill.sel.aspectoptima.aspects.checkpointable.*;
import ca.mcgill.sel.aspectoptima.aspects.checkpointing.*;
import ca.mcgill.sel.aspectoptima.aspects.context.*;
import ca.mcgill.sel.aspectoptima.aspects.copyable.*;
import ca.mcgill.sel.aspectoptima.aspects.deferrable.*;
import ca.mcgill.sel.aspectoptima.aspects.deferring.*;
import ca.mcgill.sel.aspectoptima.aspects.lockable.*;
import ca.mcgill.sel.aspectoptima.aspects.nested.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.blockable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.closable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.collaborative.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.entrysynchronizing.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.exitsynchronizing.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.openmultithreadedtransaction.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.outcomevotable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.outcomevoting.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.pausable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.shared.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.spawnsupporting.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.terminatable.*;
import ca.mcgill.sel.aspectoptima.aspects.optimisticvalidation.*;
import ca.mcgill.sel.aspectoptima.aspects.outcomeaware.*;
import ca.mcgill.sel.aspectoptima.aspects.recovering.*;
import ca.mcgill.sel.aspectoptima.aspects.traceable.*;
import ca.mcgill.sel.aspectoptima.aspects.tracing.*;
import ca.mcgill.sel.aspectoptima.aspects.transaction.*;
import ca.mcgill.sel.aspectoptima.aspects.twophaselocking.*;

public aspect AnnotationInheritanceAspect {
	/* Traceable instantiates AccessClassified */
	declare @type : @TraceableClass * : @AccessClassifiedClass;
	declare @method : (@TraceableMethod(AccessKind.READ) * *.*(..)) : @AccessClassifiedMethod(AccessKind.READ);
	declare @method : (@TraceableMethod(AccessKind.WRITE) * *.*(..)) : @AccessClassifiedMethod(AccessKind.WRITE);
	declare @method : (@TraceableMethod(AccessKind.UPDATE) * *.*(..)) : @AccessClassifiedMethod(AccessKind.UPDATE);
	
	/* Tracing instantiates Traceable */
	declare @type : @TracedClass * : @TraceableClass;
	declare @method : (@TracedMethod(AccessKind.READ) * *.*(..)) : @TraceableMethod(AccessKind.READ);
	declare @method : (@TracedMethod(AccessKind.WRITE) * *.*(..)) : @TraceableMethod(AccessKind.WRITE);
	declare @method : (@TracedMethod(AccessKind.UPDATE) * *.*(..)) : @TraceableMethod(AccessKind.UPDATE);
	
	/* Deferrable instantiates Copyable */
	declare @type : @DeferrableClass * : @CopyableClass;
	
	/* Checkpointable instantiates Copyable */
	declare @type : @CheckpointableClass * : @CopyableClass;
	
	/* Deferring instantiates Deferrable */
	declare @type : @DeferredClass * : @DeferrableClass;
	declare @method : (@DeferredMethod * *.*(..)) : @DeferrableMethod;

	/* Deferring instantiates Tracing */
	declare @type : @DeferredClass * : @TracedClass;
	declare @method : (@DeferredMethod(AccessKind.READ) * *.*(..)) : @TracedMethod(AccessKind.READ);
	declare @method : (@DeferredMethod(AccessKind.WRITE) * *.*(..)) : @TracedMethod(AccessKind.WRITE);
	declare @method : (@DeferredMethod(AccessKind.UPDATE) * *.*(..)) : @TracedMethod(AccessKind.UPDATE);
	
	/* Checkpointing instantiates Tracing */
	declare @type : @CheckpointedClass * : @TracedClass;
	declare @method : (@CheckpointedMethod(AccessKind.READ) * *.*(..)) : @TracedMethod(AccessKind.READ);
	declare @method : (@CheckpointedMethod(AccessKind.WRITE) * *.*(..)) : @TracedMethod(AccessKind.WRITE);
	declare @method : (@CheckpointedMethod(AccessKind.UPDATE) * *.*(..)) : @TracedMethod(AccessKind.UPDATE);

	/* Checkpointing instantiates Checkpointable */
	declare @type : @CheckpointedClass * : @CheckpointableClass;
	declare @method : (@CheckpointedMethod * *.*(..)) : @CheckpointableMethod;

	/* Recovering instantiates Checkpointing xor Deferring */
	declare @type : @RecoverableClass(UpdateStrategy.CHECKPOINTING) * : @CheckpointedClass;
	declare @method : (@RecoverableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.CHECKPOINTING) * *.*(..)) : @CheckpointedMethod(AccessKind.READ);
	declare @method : (@RecoverableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.CHECKPOINTING) * *.*(..)) : @CheckpointedMethod(AccessKind.WRITE);
	declare @method : (@RecoverableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.CHECKPOINTING) * *.*(..)) : @CheckpointedMethod(AccessKind.UPDATE);
	
	declare @type : @RecoverableClass(UpdateStrategy.DEFERRING) * : @DeferredClass;
	declare @method : (@RecoverableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.DEFERRING) * *.*(..)) : @DeferredMethod(AccessKind.READ);
	declare @method : (@RecoverableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.DEFERRING) * *.*(..)) : @DeferredMethod(AccessKind.WRITE);
	declare @method : (@RecoverableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.DEFERRING) * *.*(..)) : @DeferredMethod(AccessKind.UPDATE);
	
	/* TwoPhaseLocking instantiates Lockable*/
	declare @type : @TwoPhaseLockedClass * : @LockableClass;
	
	/* TwoPhaseLocking instantiates Tracing*/
	declare @type : @TwoPhaseLockedClass * : @TracedClass;
	declare @method : (@TwoPhaseLockedMethod(AccessKind.READ) * *.*(..)) : @TracedMethod(AccessKind.READ);
	declare @method : (@TwoPhaseLockedMethod(AccessKind.WRITE) * *.*(..)) : @TracedMethod(AccessKind.WRITE);
	declare @method : (@TwoPhaseLockedMethod(AccessKind.UPDATE) * *.*(..)) : @TracedMethod(AccessKind.UPDATE);

	/* Transaction instantiates Recovering */
/* TODO !!! find out why the compiler throws a NPE */
/*	declare @type : (@TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING) *) : @RecoverableClass(UpdateStrategy.CHECKPOINTING); */
	declare @method : (@TransactableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.CHECKPOINTING) * *.*(..)) : @RecoverableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.CHECKPOINTING);
	declare @method : (@TransactableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.CHECKPOINTING) * *.*(..)) : @RecoverableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.CHECKPOINTING);
	declare @method : (@TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.CHECKPOINTING) * *.*(..)) : @RecoverableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.CHECKPOINTING);

/* TODO !!! find out why the compiler throws a NPE */
/*	declare @type : @TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING) * : @RecoverableClass(UpdateStrategy.DEFERRING); */
	declare @method : (@TransactableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.DEFERRING) * *.*(..)) : @RecoverableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.DEFERRING);
	declare @method : (@TransactableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.DEFERRING) * *.*(..)) : @RecoverableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.DEFERRING);
	declare @method : (@TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.DEFERRING) * *.*(..)) : @RecoverableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.DEFERRING);
	
	/* Transaction instantiates TwoPhaseLocking xor OptimisticValidation */
	declare @type : (@TransactableClass(concurrencyControl = ConcurrencyControl.TWOPHASELOCKING) *) : @TwoPhaseLockedClass;
	declare @method : (@TransactableMethod(accessKind = AccessKind.READ, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false) * *.*(..)) : @TwoPhaseLockedMethod(AccessKind.READ);
	declare @method : (@TransactableMethod(accessKind = AccessKind.WRITE, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false) * *.*(..)) : @TwoPhaseLockedMethod(AccessKind.WRITE);
	declare @method : (@TransactableMethod(accessKind = AccessKind.UPDATE, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false) * *.*(..)) : @TwoPhaseLockedMethod(AccessKind.UPDATE);
	declare @type : (@TransactableClass(concurrencyControl = ConcurrencyControl.TWOPHASELOCKING) *) : @TwoPhaseLockedClass;
	declare @method : (@TransactableMethod(accessKind = AccessKind.READ, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true) * *.*(..)) : @TwoPhaseLockedMethod(AccessKind.READ);
	declare @method : (@TransactableMethod(accessKind = AccessKind.WRITE, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true) * *.*(..)) : @TwoPhaseLockedMethod(AccessKind.WRITE);
	declare @method : (@TransactableMethod(accessKind = AccessKind.UPDATE, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true) * *.*(..)) : @TwoPhaseLockedMethod(AccessKind.UPDATE);
	
	declare @type : @TransactableClass(concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION) * : @OptimisticallyValidatedClass;
	declare @method : (@TransactableMethod(accessKind = AccessKind.READ, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false) * *.*(..)) : @OptimisticallyValidatedMethod(AccessKind.READ);
	declare @method : (@TransactableMethod(accessKind = AccessKind.WRITE, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false) * *.*(..)) : @OptimisticallyValidatedMethod(AccessKind.WRITE);
	declare @method : (@TransactableMethod(accessKind = AccessKind.UPDATE, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false) * *.*(..)) : @OptimisticallyValidatedMethod(AccessKind.UPDATE);
	declare @type : @TransactableClass(concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION) * : @OptimisticallyValidatedClass;
	declare @method : (@TransactableMethod(accessKind = AccessKind.READ, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true) * *.*(..)) : @OptimisticallyValidatedMethod(AccessKind.READ);
	declare @method : (@TransactableMethod(accessKind = AccessKind.WRITE, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true) * *.*(..)) : @OptimisticallyValidatedMethod(AccessKind.WRITE);
	declare @method : (@TransactableMethod(accessKind = AccessKind.UPDATE, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true) * *.*(..)) : @OptimisticallyValidatedMethod(AccessKind.UPDATE);
}
