package ca.mcgill.sel.aspectoptima.aspects.copyable;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

//BEGIN implementation refinement
public interface Copyable extends Serializable {
// END implementation refinement
	// public method signatures from structural view
	Copyable copy();
	void replaceStateWith(Copyable source);
}

aspect CopyableAspect {
	// bind annotations to mandatory instantiation parameters
	declare parents : @CopyableClass * implements Copyable;
	
	// attributes & associations from structural view
	// empty
	
	// declared methods without message views
	public Copyable Copyable.copy() {
		// BEGIN implementation refinement
		return CopyableHelper.deepCopy(this);
		// END implementation refinement
	}
	
	public void Copyable.replaceStateWith(Copyable source) {
		// BEGIN implementation refinement
		/* only replace not injected fields */
		CopyableHelper.setTargetFieldsToSourceValues(source, this, false);
		// END implementation refinement
	}

	// methods shown in message views
	// empty
}

// BEGIN implementation refinement
class CopyableHelper {
	/** creates and returns a deep copy using ObjectStreams to serialize and deserialize the source */
	protected static Copyable deepCopy(Copyable source) {
		Object deepCopy = null;
		try {
			/* create a copy without injected fields before cloning it */
			ByteArrayOutputStream byteArrayOutputStream = new ByteArrayOutputStream();
			ObjectOutputStreamIgnoringInjectedFields serializer = new ObjectOutputStreamIgnoringInjectedFields(
					byteArrayOutputStream);
			serializer.writeObject(source);
			serializer.flush();
			ByteArrayInputStream memoryInputStream = new ByteArrayInputStream(
					byteArrayOutputStream.toByteArray());
			ObjectInputStream deserializer = new ObjectInputStream(
					memoryInputStream);
			deepCopy = deserializer.readObject();
			/* set injected fields to source values before returning the deepCopy */
			setTargetFieldsToSourceValues(source, deepCopy, true);
		} catch (Exception e) {
			e.printStackTrace();
		}
		return (Copyable) deepCopy;
	}
	
	/* TODO extend Copyable to allow fields to be annotated with @omitCopying or something similar */
	protected static Object createCopyWithoutInjectedFields(Object source) {
		Object shallowCopy = null;
		try {
			Class<?> sourceClass = source.getClass();
			shallowCopy = sourceClass.newInstance();
			/* get all fields of the class */
			List<Field> fields = new ArrayList<Field>();
			CopyableHelper.addDeclaredAndInheritedFields(sourceClass, fields);
			/* make protected and private fields accessible*/
			Field[] fieldsArray = new Field[fields.size()];
			fieldsArray = fields.toArray(fieldsArray);
			AccessibleObject.setAccessible(fieldsArray, true);
			for (Field field : fields) {
				if (!Modifier.isFinal(field.getModifiers())) {
					if (field.getName().startsWith("ajc")) {
						/* omit injected field */
						field.set(shallowCopy, null);
					}
					else {
						/* point to the source value */
						field.set(shallowCopy, field.get(source));
					}
				}
			}
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		} catch (InstantiationException e) {
			e.printStackTrace();
		}
		return shallowCopy;
	}

	/** sets fields of the target to the corresponding fields of the source <br>
	 * only operates on injected fields if injectedFields is true <br>
	 * only operates on not injected fields if injectedFields is false */
	protected static void setTargetFieldsToSourceValues(Object source,
			Object target, boolean injectedFields) {
		if (source != null) {
			Class<?> sourceClass = source.getClass();
			/* get all fields of the class */
			List<Field> fields = new ArrayList<Field>();
			CopyableHelper.addDeclaredAndInheritedFields(sourceClass, fields);
			/* make protected and private fields accessible */
			Field[] fieldsArray = new Field[fields.size()];
			fieldsArray = fields.toArray(fieldsArray);
			AccessibleObject.setAccessible(fieldsArray, true);
			for (Field field : fields) {
				if (!Modifier.isFinal(field.getModifiers())) {
					if(injectedFields) {
						/* skip fields that are not injected */
						if (!field.getName().startsWith("ajc")) {
							try {
								Object fieldSourceValue = field.get(source);
								if (fieldSourceValue != null) {
									Class<?> fieldValueClass = fieldSourceValue.getClass();
									/* not injected non-primitive fields might contain injected fields*/
									if (!CopyableHelper.isWrapperType(fieldValueClass)) {
										Object fieldTargetValue = field.get(target);
										setTargetFieldsToSourceValues(fieldSourceValue, fieldTargetValue, injectedFields);
									}
								}
							} catch (IllegalAccessException e) {
								e.printStackTrace();
							}
							/* avoid setting of not injected fields */
							continue;
						}
					}
					else {
						/* skip injected fields */
						if (field.getName().startsWith("ajc")) {
							continue;
						}
					}
					try {
						/* point to the source value */
						field.set(target, field.get(source));
					} catch (IllegalAccessException e) {
						e.printStackTrace();
					}
				}
			}
			/* if the source is an array consider each entry */
			if (source.getClass().isArray()) {
				Object[] sourceArray = (Object[]) source;
				Object[] targetArray = (Object[]) target;
				for (int i = 0; i < sourceArray.length; i++) {
					setTargetFieldsToSourceValues(sourceArray[i], targetArray[i], injectedFields);
				}
			}
		}
	}

	protected static void addDeclaredAndInheritedFields(Class<?> c,
			List<Field> fields) {
		fields.addAll(Arrays.asList(c.getDeclaredFields()));
		Class<?> superClass = c.getSuperclass();
		if (superClass != null) {
			addDeclaredAndInheritedFields(superClass, fields);
		}
	}
	
	private static final Set<Class<?>> WRAPPER_TYPES = new HashSet<Class<?>>(Arrays.asList(
			Boolean.class, Character.class, Byte.class, Short.class,
			Integer.class, Long.class, Float.class, Double.class, Void.class));

	protected static boolean isWrapperType(Class<?> clazz) {
		return WRAPPER_TYPES.contains(clazz);
	}
}

class ObjectOutputStreamIgnoringInjectedFields extends ObjectOutputStream {
	protected ObjectOutputStreamIgnoringInjectedFields(OutputStream out)
			throws IOException, SecurityException {
		super(out);
		enableReplaceObject(true);
	}
	
	@Override
	protected Object replaceObject(Object obj) throws IOException {
		if (CopyableHelper.isWrapperType(obj.getClass())) {
			return obj;
		}
		else {
			return CopyableHelper.createCopyWithoutInjectedFields(obj);
		}
	}
}
// END implementation refinement