package ca.mcgill.sel.aspectoptima.aspects.deferrable;

import java.util.HashMap;
import java.util.Map;

import ca.mcgill.sel.aspectoptima.aspects.context.Context;
import ca.mcgill.sel.aspectoptima.aspects.copyable.Copyable;

public interface Deferrable extends Copyable {
	// public method signatures from structural view
	/** returns the version that is mapped to the provided context or creates, maps and returns a new version if no version is mapped yet */
	Deferrable getVersion(Context context);
	/** returns the original Deferrable of which this Deferrable is a copy */
	Deferrable getOriginal();
	void setOriginal(Deferrable original);
	void updateOriginal(Context context);
}

aspect DeferrableAspect {
	// bind annotations to mandatory instantiation parameters
	declare parents : @DeferrableClass * implements Deferrable;
	
	// attributes & associations from structural view
	private Map<Context, Deferrable> Deferrable.contextToVersionMap = new HashMap<Context, Deferrable>();
	private Deferrable Deferrable.original = null;
	
	// BEGIN implementation refinement
	after(Deferrable deferrable) : initialization(Deferrable.new(..)) && this(deferrable){
		deferrable.original = deferrable;
	}
	// END implementation refinement
	
	// declared methods without message views
	public Deferrable Deferrable.getOriginal() {
		// auto-generated getter implementation
		return this.original;
	}
	
	public void Deferrable.setOriginal(Deferrable original) {
		// auto-generated setter implementation
		this.original = original;
	}

	// methods shown in message views
	public Deferrable Deferrable.getVersion(Context context) {
		Deferrable version;
		Deferrable original = this.getOriginal();
		if (original != this) {
			version = this;
		}
		else {
			version = this.contextToVersionMap.get(context);
			if (version == null) {
				version = (Deferrable) original.copy();
				this.contextToVersionMap.put(context, version);
			}
		}
		return version;
	}
	
	/** retrieves the version for the provided context, replaces the original state with the version and deletes the version */
	public void Deferrable.updateOriginal(Context context) {
		Deferrable original = this.getOriginal();
		Deferrable version = this.getVersion(context);
		original.replaceStateWith(version);
		original.deleteVersion(context);
	}
	
	/** deletes the version that is mapped to the provided context */
	void Deferrable.deleteVersion(Context context) {
		this.contextToVersionMap.remove(context);
	}
}
