package ca.mcgill.sel.aspectoptima.aspects.deferring;

import org.aspectj.lang.annotation.AdviceName;

import ca.mcgill.sel.aspectoptima.aspects.deferrable.Deferrable;
import ca.mcgill.sel.aspectoptima.aspects.tracing.Traced;
import ca.mcgill.sel.aspectoptima.interfacing.ParticipantPool;

public interface Deferred extends Deferrable, Traced {
	// public method signatures from structural view
	// empty
}

aspect DeferredAspect {
	// bind annotations to mandatory instantiation parameters
	declare parents : @DeferredClass * implements Deferred;
	
	// attributes & associations from structural view
	// empty
	
	// declared methods without message views
	// empty

	// methods shown in message views
	/** defers each call to a deferred method by performing it on behalf 
	 * of its context on a separate version for the context <br>
	 * @return the Object returned by the advised method */
	@AdviceName("deferMethod")
	Object around(Deferred target): execution(@DeferredMethod * *.*(..)) && target(target) {
		Deferrable original = target.getOriginal();
		/* check whether we are already calling a method on a version */
		if (original != target) {
			/* execute the method directly on the target as it is already a version */
			return proceed(target);
		}
		else {
			DeferringParticipant participant = (DeferringParticipant) DeferringParticipantStatic.getCurrent();
			DeferringContext context = (DeferringContext) participant.getContext();
			Deferred version = (Deferred) target.getVersion(context);
			/* execute the method on the separate version */
			return proceed(version);
		}
	}
}
