package ca.mcgill.sel.aspectoptima.aspects.deferring;

import java.lang.annotation.Documented;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;
import java.lang.annotation.ElementType;
import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;

/**
 * Marks a method as Deferred for the AspectOPTIMA Framework<br>
 * The parameter value() should provide the kind of access (i.e. AccessKind.READ,
 * AccessKind.WRITE, AccessKind.UPDATE).<br>
 * The syntactically equivalent Annotation @TracingMethod is added automatically.<br>
 * Objects containing Deferred methods should be annotated with @DeferredObject.
 */
@Target( { ElementType.METHOD })
@Documented
@Retention(RetentionPolicy.RUNTIME)
public @interface DeferredMethod {
	AccessKind value();
}