package ca.mcgill.sel.aspectoptima.aspects.optimisticvalidation;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;

/**
 * Marks a method as OptimisticallyValidated for the AspectOPTIMA Framework<br>
 * The parameter value() should provide the kind of access (i.e.
 * AccessKind.READ, AccessKind.WRITE, AccessKind.UPDATE).<br>
 * The syntactically equivalent Annotation @TracedMethod is added automatically.<br>
 * Objects containing TwoPhaseLocked methods should be annotated with
 * @OptimisticallyValidatedClass.
 */
@Target( { ElementType.METHOD })
@Documented
@Retention(RetentionPolicy.RUNTIME)
public @interface OptimisticallyValidatedMethod {
	AccessKind value();
}