package ca.mcgill.sel.aspectoptima.aspects.recovering;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

import ca.mcgill.sel.aspectoptima.aspects.recovering.UpdateStrategy;

/**
 * Marks an object as Recoverable for the AspectOPTIMA Framework. <br>
 * The parameter value() should provide the update strategy (i.e
 * UpdateStrategy.CHECKPOINTING or UpdateStrategy.DEFERRING). <br>
 * Methods of Recoverable objects can be marked as Recovering using
 * @RecoveringCheckpointed. <br>
 * The corresponding annotations @CheckpointedClass and @DeferredObject will be
 * added automatically. <br>
 * Recoverable objects may contain methods of both recovery mechanisms.
 */
@Target( { ElementType.TYPE })
@Documented
@Retention(RetentionPolicy.RUNTIME)
public @interface RecoverableClass {
	UpdateStrategy value();
}