package ca.mcgill.sel.aspectoptima.aspects.recovering;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;
import ca.mcgill.sel.aspectoptima.aspects.recovering.UpdateStrategy;

/**
 * Marks a method as Recovering for the AspectOPTIMA Framework using
 * Checkpointing as recovery mechanism.<br>
 * The parameter accessKind() should provide the kind of access (i.e.
 * AccessKind.READ, AccessKind.WRITE, AccessKind.UPDATE).<br>
 * The parameter value() should provide the update strategy (i.e
 * UpdateStrategy.CHECKPOINTING or UpdateStrategy.DEFERRING). <br>
 * The syntactically equivalent Annotation @CheckpointedMethod is added
 * automatically.<br>
 * Objects containing Recovering methods should be annotated with
 * @RecoverableClass.
 */
@Target( { ElementType.METHOD })
@Documented
@Retention(RetentionPolicy.RUNTIME)
public @interface RecoverableMethod {
	AccessKind accessKind();
	UpdateStrategy updateStrategy();
}