package ca.mcgill.sel.aspectoptima.aspects.transaction;

import org.aspectj.lang.annotation.AdviceName;

import ca.mcgill.sel.aspectoptima.aspects.recovering.Recoverable;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableMethod;
import ca.mcgill.sel.aspectoptima.interfacing.AspectOPTIMAInternal;

public interface Transactable extends Recoverable {
	// public method signatures from structural view
	// empty
}

aspect TransactableAspect {
	// attributes & associations from structural view
	// empty
	
	// declared methods without message views
	// empty

	// methods shown in message views
	/** surrounds every call to a transacted method with "begin" and "commit" 
	 * if they are not already called within an existing transaction */
	@AdviceName("transactMethod")
	// BEGIN implementation refinement
	Object around(TransactableMethod transactedMethodAnnotation): execution(* *.*(..)) && @annotation(transactedMethodAnnotation) {
		TransactionParticipant participant = AspectOPTIMAInternal.getTransactionParticipant(transactedMethodAnnotation);
		// END implementation refinement
		boolean alreadyTransacted = participant.isAlreadyTransacted();
		if (!alreadyTransacted) {
			participant.beginTransaction();
		}
		Object returned = proceed(transactedMethodAnnotation);
		if (!alreadyTransacted) {
			participant.commitTransaction();
		}
		return returned;
	}
}
