package ca.mcgill.sel.aspectoptima.aspects.twophaselocking;

import java.lang.annotation.Documented;
import java.lang.annotation.ElementType;
import java.lang.annotation.Retention;
import java.lang.annotation.RetentionPolicy;
import java.lang.annotation.Target;

/**
 * Marks an object as TwoPhaseLocked for the AspectOPTIMA Framework. <br>
 * The parameter value() should provide a Lock class that extends
 * java.util.concurrent.locks.ReadWriteLock and defaults to
 * java.util.concurrent.locks.ReentrantReadWriteLock.class
 * The needed Annotations @TracedClass and @LockableObject are added automatically. <br>
 * Methods of TwoPhaseLocked objects can be marked as TwoPhaseLocked using @TwoPhaseLockedMethod. <br>
 */
@Target({ElementType.TYPE}) 
@Documented
@Retention(RetentionPolicy.RUNTIME)
public @interface TwoPhaseLockedClass {
	Class<? extends java.util.concurrent.locks.ReadWriteLock> value() default java.util.concurrent.locks.ReentrantReadWriteLock.class;
}
