package ca.mcgill.sel.aspectoptima.aspects.twophaselocking;

import java.util.HashMap;
import ca.mcgill.sel.aspectoptima.aspects.context.Participant;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;
import ca.mcgill.sel.aspectoptima.aspects.lockable.LockableForTwoPhaseLocking;
import ca.mcgill.sel.aspectoptima.aspects.traceable.Trace;
import ca.mcgill.sel.aspectoptima.aspects.traceable.Traceable;
import ca.mcgill.sel.aspectoptima.aspects.tracing.TracingParticipant;

public interface TwoPhaseLockingParticipant extends TracingParticipant {
	// public method signatures from structural view
	// empty
}

aspect TwoPhaseLockingParticipantAspect {
	// bind annotations to mandatory instantiation parameters
	declare parents : @TwoPhaseLockingParticipantClass * implements TwoPhaseLockingParticipant;
	
	// attributes & associations from structural view
	// empty
	// BEGIN implementation refinement
	private Map<Traceable, Set<AccessKind>> TwoPhaseLockingParticipant.targetToReleasedLockKindsMap = new HashMap<Traceable, Set<AccessKind>>();
	// END implementation refinement
	
	// declared methods without message views
	private void TwoPhaseLockingParticipant.releaseLockIfPresent(AccessKind accessKind, Traceable target) {
		// BEGIN implementation refinement
		Set<AccessKind> releasedLockKinds = this.targetToReleasedLockKindsMap.get(target);
		if (releasedLockKinds == null) {
			releasedLockKinds = new HashSet<AccessKind>();
			targetToReleasedLockKindsMap.put(target, releasedLockKinds);
		}
		if (accessKind == AccessKind.UPDATE) {
			accessKind = AccessKind.WRITE;
		}
		if (!releasedLockKinds.contains(accessKind)) {
			((LockableForTwoPhaseLocking) target).releaseLock(accessKind);
			releasedLockKinds.add(accessKind);
		}
		// END implementation refinement
	}

	// methods shown in message views
	void around(TwoPhaseLockingParticipant target) : execution(void Participant.leaveContext()) && target(target) {
		TwoPhaseLockedContext oldContext = (TwoPhaseLockedContext) target.getContext();
		proceed(target);
		target.releaseAcquiredLocks(oldContext);
	}
	
	private void TwoPhaseLockingParticipant.releaseAcquiredLocks(TwoPhaseLockedContext c) {
		List<Trace> traces = c.getTraces();
		for (Trace t :traces) {
			Traceable traceTarget = t.getTarget();
			AccessKind accessKind = t.getAccessKind();
			this.releaseLockIfPresent(accessKind, traceTarget);
		}
	}
}
