package ca.mcgill.sel.aspectoptima.conflictresolutions.deferrablenested;
import ca.mcgill.sel.aspectoptima.aspects.context.Context;
import ca.mcgill.sel.aspectoptima.aspects.deferrable.Deferrable;
import ca.mcgill.sel.aspectoptima.aspects.nested.NestedContext;

privileged aspect DeferrableAspect {
	// attributes & associations from structural view
	// empty
	
	// declared methods without message views
	// empty

	// methods shown in message views
	Deferrable around(Deferrable target, NestedContext initialContext) : execution(Deferrable Deferrable.getVersion(..)) && target(target) && args(initialContext) {
		NestedContext currentContext = initialContext;
		Deferrable version = target.contextToVersionMap.get(initialContext);
		/* move upwards in the hierarchy until a version of a ancestor is found */
		/* if a new version is created somewhere we will start at the bottom of the hierarchy again */
		while (version == null) {
			NestedContext parentContext = currentContext.getParent();
			if (parentContext == null) {
				/* if no parentContext exists copy the original object for the currentContext */
				target.copyAndMapNewVersion(target.original, currentContext);
				/* start context retrieval again from the initial context */
				parentContext = initialContext;
			}
			else {
				/* parentContext != null */
				Deferrable parentVersion = target.contextToVersionMap.get(parentContext);
				if (parentVersion != null) {
					/* if the parentContext has a version copy it for the currentContext */
					target.copyAndMapNewVersion(parentVersion, currentContext);
					/* start context retrieval again from the initial context */
					parentContext = initialContext;
				}
			}
			/* either parentContext = currentContext.getParent() or parentContext = initialContext */
			version = target.contextToVersionMap.get(parentContext);
			/* => climb up the hierarchy or start again */
			currentContext = parentContext;
		}
		return version;	
	}
	
	private void Deferrable.copyAndMapNewVersion(Deferrable source, Context context) {
		Deferrable newVersion = (Deferrable) source.copy();
		/* adjust the newVersion's pointer to the original 
		 * point to the parentalVersion in order to update the right original in case of a commit */
		newVersion.setOriginal(source);
		this.contextToVersionMap.put(context, newVersion);
	}
}