package ca.mcgill.sel.aspectoptima.interfacing;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;
import ca.mcgill.sel.aspectoptima.aspects.context.Context;
import ca.mcgill.sel.aspectoptima.aspects.recovering.UpdateStrategy;
import ca.mcgill.sel.aspectoptima.aspects.transaction.ConcurrencyControl;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableMethod;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactionParticipant;

public class AspectOPTIMAInternal {
	@SuppressWarnings("unchecked")
	public static TransactionParticipant getTransactionParticipant(String updateStrategy, String concurrencyControl, boolean nested) throws ClassNotFoundException {
		String variabilityEncoding = getVariabilityEncoding(updateStrategy, concurrencyControl, nested);
		String participantImplClassName = TransactionParticipant.class.getName() + variabilityEncoding + "Impl";
		Class<? extends TransactionParticipant> participantImplClass = (Class<? extends TransactionParticipant>) Class.forName(participantImplClassName);
		return ParticipantPool.getParticipant(participantImplClass);	
	}

	public static TransactionParticipant getTransactionParticipant(TransactableMethod transactedMethodAnnotation) {
		try {
			Class<? extends TransactionParticipant> participantImplClass = getTransactionParticipantClass(transactedMethodAnnotation);
			return ParticipantPool.getParticipant(participantImplClass);
		} catch (ClassNotFoundException e) {
			e.printStackTrace();
		}
		return null;
	}

	@SuppressWarnings("unchecked")
	private static Class<? extends TransactionParticipant> getTransactionParticipantClass(TransactableMethod transactedMethodAnnotation) throws ClassNotFoundException {
		String variabilityEncoding = getVariabilityEncoding(transactedMethodAnnotation);
		String participantImplClassName = TransactionParticipant.class.getName() + variabilityEncoding;
		return (Class<? extends TransactionParticipant>) Class.forName(participantImplClassName);
	}
	
	public static String getVariabilityEncoding(String updateStrategy, String concurrencyControl, boolean nested) {
		UpdateStrategy updateStrategyVal = UpdateStrategyForString(updateStrategy);
		ConcurrencyControl concurrencyControlVal = ConcurrencyControlForString(concurrencyControl);
		String nestedString = nested ? "Nested" : "";
		return updateStrategyVal.toString() + concurrencyControlVal.toString() + nestedString;
	}
	
	public static AccessKind AccessKindForString(String s) {
		AccessKind result = AccessKind.UPDATE;
		try {
			s = s.toUpperCase();
			result = AccessKind.valueOf(s);
		} catch (Exception e) {
			/* swallow */
		}
		return result;
	}

	public static UpdateStrategy UpdateStrategyForString(String s) {
		UpdateStrategy result = UpdateStrategy.CHECKPOINTING;
		try {
			s = s.toUpperCase();
			result = UpdateStrategy.valueOf(s);
		} catch (Exception e) {
			/* swallow */
		}
		return result;
	}
	
	public static ConcurrencyControl ConcurrencyControlForString(String s) {
		ConcurrencyControl result = ConcurrencyControl.TWOPHASELOCKING;
		try {
			s = s.toUpperCase();
			result = ConcurrencyControl.valueOf(s);
		} catch (Exception e) {
			/* swallow */
		}
		return result;
	}
	
	public static String getVariabilityEncoding(TransactableMethod transactedMethodAnnotation) {
		String updateStrategy = transactedMethodAnnotation.updateStrategy().toString();
		String concurrencyControl = transactedMethodAnnotation.concurrencyControl().toString();
		String nested = transactedMethodAnnotation.nested() ? "Nested" : "";
		return updateStrategy + concurrencyControl + nested;
	}
}
