package ca.mcgill.sel.aspectoptima.interfacing;

import org.aspectj.lang.annotation.AdviceName;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.*;
import ca.mcgill.sel.aspectoptima.aspects.checkpointable.*;
import ca.mcgill.sel.aspectoptima.aspects.checkpointing.*;
import ca.mcgill.sel.aspectoptima.aspects.context.*;
import ca.mcgill.sel.aspectoptima.aspects.copyable.*;
import ca.mcgill.sel.aspectoptima.aspects.deferrable.*;
import ca.mcgill.sel.aspectoptima.aspects.deferring.*;
import ca.mcgill.sel.aspectoptima.aspects.lockable.*;
import ca.mcgill.sel.aspectoptima.aspects.nested.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.blockable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.closable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.collaborative.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.entrysynchronizing.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.exitsynchronizing.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.openmultithreadedtransaction.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.outcomevotable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.outcomevoting.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.pausable.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.shared.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.spawnsupporting.*;
import ca.mcgill.sel.aspectoptimaomtt.aspects.terminatable.*;
import ca.mcgill.sel.aspectoptima.aspects.optimisticvalidation.*;
import ca.mcgill.sel.aspectoptima.aspects.outcomeaware.*;
import ca.mcgill.sel.aspectoptima.aspects.recovering.*;
import ca.mcgill.sel.aspectoptima.aspects.traceable.*;
import ca.mcgill.sel.aspectoptima.aspects.tracing.*;
import ca.mcgill.sel.aspectoptima.aspects.transaction.*;
import ca.mcgill.sel.aspectoptima.aspects.twophaselocking.*;
import ca.mcgill.sel.aspectoptima.interfacing.*;

public aspect InterfacingAspect {
	declare precedence : InterfacingAspect, *;
	
	/** add annotations for class automatically based on the annotation of their methods */
	declare @type : hasmethod(@AccessClassifiedMethod * *.*(..)) : @AccessClassifiedClass;
	
	declare @type : hasmethod(@TraceableMethod * *.*(..)) : @TraceableClass;
	
	declare @type : hasmethod(@TracedMethod * *.*(..)) : @TracedClass;
	
	declare @type : hasmethod(@DeferrableMethod * *.*(..)) : @DeferrableClass;
	
	declare @type : hasmethod(@DeferredMethod * *.*(..)) : @DeferredClass;
	
	declare @type : hasmethod(@CheckpointableMethod * *.*(..)) : @CheckpointableClass;
	
	declare @type : hasmethod(@CheckpointedMethod * *.*(..)) : @CheckpointedClass;
	
	declare @type : hasmethod(@RecoverableMethod(updateStrategy = UpdateStrategy.CHECKPOINTING) * *.*(..)) : @CheckpointedClass;
	
	declare @type : hasmethod(@RecoverableMethod(updateStrategy = UpdateStrategy.DEFERRING) * *.*(..)) : @DeferredClass;
	
	declare @type : hasmethod(@TransactableMethod(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING) * *.*(..)) : @TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING);
	
	declare @type : hasmethod(@TransactableMethod(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING) * *.*(..)) : @TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING);

	declare @type : hasmethod(@TransactableMethod(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION) * *.*(..)) : @TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION);
	
	declare @type : hasmethod(@TransactableMethod(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION) * *.*(..)) : @TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION);
	
	/** add annotations for auto transacted class */
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false) *) : @TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING);
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true) *) : @TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING);
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false) *) : @TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING);
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true) *) : @TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING);
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false) *) : @TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION);
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true) *) : @TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION);
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false) *) : @TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION);
	declare @type : (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true) *) : @TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION);
	
	/** add annotations for methods of auto transacted class */
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false);
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true);
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false);
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true);
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false);
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true);
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = false);
	declare @method : (* (@AutoTransactedClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true) *).*(..)) : @TransactableMethod(accessKind = AccessKind.UPDATE, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.OPTIMISTICVALIDATION, nested = true);	

	pointcut abortOnThrowingMethodExecution(AbortTransactionOnThrowing abortAnnotation, TransactableMethod transactedMethodAnnotation) : execution(* *.*(..)) && @annotation(abortAnnotation) && @annotation(transactedMethodAnnotation);

	/**
	 * aborts the current transaction if a method annotated with 
	 * @AbortTransactionOnThrowing(ThrowableToCatch.class) throws a Throwable
	 */
	@AdviceName("abortTransactionOnThrowing")
	after(AbortTransactionOnThrowing abortAnnotation, TransactableMethod transactedMethodAnnotation) throwing(Throwable t): abortOnThrowingMethodExecution(abortAnnotation, transactedMethodAnnotation) {
		Class<?> abortClass = abortAnnotation.value();
		/* check whether: t instanceof abortClass */
		if (abortClass.isInstance(t)) {
			TransactionParticipant participant = AspectOPTIMAInternal.getTransactionParticipant(transactedMethodAnnotation);
			participant.abortTransaction();
		}
	}
	
	declare parents : java.lang.Thread implements OMTTParticipantCheckpointingTwoPhaseLockingNested;
}
