package ca.mcgill.sel.aspectoptima.interfacing;

import java.lang.reflect.Constructor;

import ca.mcgill.sel.aspectoptima.aspects.context.Participant;
import ca.mcgill.sel.aspectoptimaomtt.aspects.openmultithreadedtransaction.OMTTParticipant;
import ca.mcgill.sel.aspectoptima.aspects.outcomeaware.BinaryOutcomeImpl;
import ca.mcgill.sel.aspectoptima.aspects.outcomeaware.OutcomeControllingParticipant;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactionParticipant;


public class ParticipantPool {
	private static ThreadLocalParticipant threadLocalParticipant = new ThreadLocalParticipant();

	public static Participant getThreadLocalParticipant() {
		return threadLocalParticipant.get();
	}

	public static void setThreadLocalParticipant(Participant participant) {
		ParticipantPool.threadLocalParticipant.set(participant);
	}

	@SuppressWarnings("unchecked")
	public static <ParticipantT extends Participant> ParticipantT getParticipant(
			Class<? extends ParticipantT> participantClass) {
		try {

			Participant currentLocalParticipant = threadLocalParticipant.get();
			if (currentLocalParticipant != null) {
				if (participantClass.isInstance(currentLocalParticipant)) {
					return (ParticipantT) currentLocalParticipant;
				}
			}
			/* if the current participant is not compatible to the requested
			 * participant set the current participant accordingly
			 * FIXME throw exception if the context of the currentParticipant is
			 * not null */
			String requestedClassName = participantClass.getName();
			boolean testing = (!requestedClassName.startsWith(TransactionParticipant.class.getName()) && !requestedClassName.startsWith(OMTTParticipant.class.getName()));
			if (testing) {
				String simpleName = participantClass.getSimpleName();
				if (requestedClassName.startsWith("ca.mcgill.sel.aspectoptima.")) {
					requestedClassName = "ca.mcgill.sel.aspectoptima.test." + simpleName;
				}
				else {
					requestedClassName = "ca.mcgill.sel.aspectoptimaomtt.test." + simpleName;
				}
			}
			String implementationClassName = requestedClassName + "Impl";
			Class<? extends Participant> implementationClass = (Class<? extends Participant>) Class.forName(implementationClassName);
			Constructor<? extends Participant> constructor = implementationClass.getDeclaredConstructor((Class<?>[]) null);
			constructor.setAccessible(true);
			Participant newParticipant = constructor.newInstance((Object[]) null);
			threadLocalParticipant.set(newParticipant);
			return (ParticipantT) newParticipant;
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}

	@SuppressWarnings("unchecked")
	public static <ParticipantT extends Participant> ParticipantT claimParticipant(
			Class<ParticipantT> c) {
		/* FIXME throw exception if cast or instantiation fail
		 * e.g. when a participant is requested that is incompatible to the
		 * current participant */
		ParticipantT currentLocalParticipant = (ParticipantT) threadLocalParticipant
				.get();
		if (currentLocalParticipant == null) {
			/* FIXME throw an exception if a participant is claimed but not set */
		}
		return currentLocalParticipant;
	}

	public static Participant claimParticipant() {
		Participant currentLocalParticipant = (Participant) threadLocalParticipant
				.get();
		if (currentLocalParticipant == null) {
			/* FIXME throw an exception if a participant is claimed but not set */
		}
		return currentLocalParticipant;
	}

	public static <ParticipantT extends Participant> void begin(
			Class<ParticipantT> c) {
		ParticipantT participant = getParticipant(c);
		participant.createAndEnterContext();
	}

	protected static <ParticipantT extends Participant> void end(
			Class<ParticipantT> c) {
		ParticipantT participant = claimParticipant(c);
		participant.leaveContext();
	}

	public static void end() {
		Participant participant = claimParticipant();
		participant.leaveContext();
	}

	public static void abort() {
		/* FIXME throw exception if currentParticipant !instanceof
		 * OutcomeControllingParticipant */
		OutcomeControllingParticipant participant = (OutcomeControllingParticipant) claimParticipant();
		participant.voteAndLeaveContext(new BinaryOutcomeImpl(false));
	}

	public static void commit() {
		/* FIXME throw exception if currentParticipant !instanceof
		 * OutcomeControllingParticipant */
		OutcomeControllingParticipant participant = (OutcomeControllingParticipant) claimParticipant();
		participant.voteAndLeaveContext(new BinaryOutcomeImpl(true));
	}
}

class ThreadLocalParticipant extends ThreadLocal<Participant> {
	@Override
	protected Participant initialValue() {
		return null;
	}
}