package ca.mcgill.sel.aspectoptima.test;

import org.junit.Test;
import static org.junit.Assert.assertEquals;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;
import ca.mcgill.sel.aspectoptima.aspects.transaction.ConcurrencyControl;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableClass;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableMethod;
import ca.mcgill.sel.aspectoptima.aspects.recovering.RecoverableClass;
import ca.mcgill.sel.aspectoptima.aspects.recovering.UpdateStrategy;

@TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING)
public class NestedTransaction2PLAdvancedDeferringTest {
	NestedTransaction2PLAdvancedDeferringTest child;
	int val = 0;

	@TransactableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true)
	void setVal(int val) {
		this.val = val;
	}
	
	@TransactableMethod(accessKind = AccessKind.WRITE, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true)
	void setValOfChild(int val) {
		child.val = val;
	}
	
	@Test
	public void deferringNestedNoAcNestAcCommitAbort() {
		NestedTransaction2PLAdvancedDeferringTest t = new NestedTransaction2PLAdvancedDeferringTest();
		t.val = 1;
		
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		t.setVal(2);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(t.val, 1);
		t.setVal(3);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(t.val, 1);
	}
	
	@Test
	public void deferringNestedNoAcNestAcCommitCommit() {
		NestedTransaction2PLAdvancedDeferringTest t = new NestedTransaction2PLAdvancedDeferringTest();
		t.val = 1;
		
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		t.setVal(2);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(t.val, 1);
		t.setVal(3);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(t.val, 3);
	}
	
	@Test
	public void deferringNestedNoAcNestAcAbortCommit() {
		NestedTransaction2PLAdvancedDeferringTest t = new NestedTransaction2PLAdvancedDeferringTest();
		t.val = 1;
		
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		t.setVal(2);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(t.val, 1);
		t.setVal(3);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(t.val, 3);
	}
	
	@Test
	public void deferringNestedNoAcNestAcAbortAbort() {
		NestedTransaction2PLAdvancedDeferringTest t = new NestedTransaction2PLAdvancedDeferringTest();
		t.val = 1;
		
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLNestedTransaction();
		t.setVal(2);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(t.val, 1);
		t.setVal(3);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(t.val, 1);
	}
}