package ca.mcgill.sel.aspectoptima.test;

import static org.junit.Assert.assertEquals;

import org.junit.Test;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;
import ca.mcgill.sel.aspectoptima.aspects.transaction.ConcurrencyControl;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableClass;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableMethod;
import ca.mcgill.sel.aspectoptima.aspects.recovering.UpdateStrategy;

@TransactableClass(updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING)
public class NestedTransaction2PLTest {
	int insideTest = 0;

	@Test
	public void checkpointedBeginBeginAbortCommitTest() {
		NestedTransaction2PLTest test = new NestedTransaction2PLTest();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.foo(1);
		assertEquals(2, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.bar(4, 5);
		assertEquals(4, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(2, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(2, test.insideTest);
	}

	@Test
	public void checkpointedBeginBeginAbortAbortTest() {
		NestedTransaction2PLTest test = new NestedTransaction2PLTest();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.foo(1);
		assertEquals(2, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.bar(4, 5);
		assertEquals(4, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(2, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(0, test.insideTest);
	}

	@Test
	public void checkpointedBeginBeginCommitAbortTest() {
		NestedTransaction2PLTest test = new NestedTransaction2PLTest();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.foo(1);
		assertEquals(2, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.bar(4, 5);
		assertEquals(4, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(4, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(0, test.insideTest);
	}

	@Test
	public void checkpointedBeginBeginCommitCommitTest() {
		NestedTransaction2PLTest test = new NestedTransaction2PLTest();
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.foo(1);
		assertEquals(2, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginCheckpointed2PLNestedTransaction();
		test.bar(4, 5);
		assertEquals(4, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(4, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(4, test.insideTest);
	}

	@TransactableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true)
	void foo(int a) {
		insideTest = a;
		bar(2, 3);
	}

	@TransactableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.CHECKPOINTING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = true)
	void bar(int a, int b) {
		insideTest = a;
	}
}