package ca.mcgill.sel.aspectoptima.test;

import static org.junit.Assert.assertEquals;

import org.junit.Test;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;
import ca.mcgill.sel.aspectoptima.aspects.transaction.ConcurrencyControl;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableClass;
import ca.mcgill.sel.aspectoptima.aspects.transaction.TransactableMethod;
import ca.mcgill.sel.aspectoptima.aspects.recovering.UpdateStrategy;

@TransactableClass(updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING)
public class Transaction2PLTestDeferring {
	int insideTest = 0;

	@Test
	public void transactionDeferred2PLAbortTest() {
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLTransaction();
		Transaction2PLTestDeferring test = new Transaction2PLTestDeferring();
		test.foo2(1);
		test.bar2(4, 5);
		assertEquals(0, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.abort();
		assertEquals(0, test.insideTest);
	}

	@TransactableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false)
	void foo2(int a) {
		insideTest = a;
		bar2(2, 3);
	}

	@TransactableMethod(accessKind = AccessKind.READ, updateStrategy = UpdateStrategy.DEFERRING, concurrencyControl = ConcurrencyControl.TWOPHASELOCKING, nested = false)
	void bar2(int a, int b) {
		insideTest = a;
	}

	@Test
	public void transactionDeferred2PLCommitTest() {
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.beginDeferred2PLTransaction();
		Transaction2PLTestDeferring test = new Transaction2PLTestDeferring();
		test.foo2(1);
		test.bar2(4, 5);
		assertEquals(0, test.insideTest);
		ca.mcgill.sel.aspectoptima.test.AspectOPTIMATesting.commit();
		assertEquals(4, test.insideTest);
	}

	@Test
	public void transactionDeferred2PLAutoTest() {
		Transaction2PLTestDeferring test = new Transaction2PLTestDeferring();
		test.foo2(1);
		assertEquals(2, test.insideTest);
		test.bar2(4, 5);
		assertEquals(4, test.insideTest);
	}
}
