package ca.mcgill.sel.aspectoptimaomtt.aspects.shared;

import java.lang.reflect.Method;

import org.aspectj.lang.annotation.AdviceName;
import org.aspectj.lang.reflect.MethodSignature;

import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessClassified;
import ca.mcgill.sel.aspectoptima.aspects.accessclassified.AccessKind;
import ca.mcgill.sel.aspectoptima.aspects.lockable.LockableForShared;

public interface Shared extends AccessClassified, LockableForShared {
	// public method signatures from structural view
	// empty
}

aspect SharedAspect {
	// bind annotations to mandatory instantiation parameter interfaces
	declare parents : @SharedClass * implements Shared;
		
	// attributes & associations from structural view
	// empty
	
	// declared methods without message views
	// empty
	
	// methods shown in message views
	/** protects each call to a shared method by acquiring a lock 
	 * that corresponds to the accessKind of the method and
	 * releases the lock after proceeding with the advised method
	 * @return the Object returned by the advised method */
	@AdviceName("shareMethod")
	Object around(Shared target): execution(@SharedMethod * *.*(..)) && target(target) {
		Method m = ((MethodSignature) thisJoinPointStaticPart.getSignature()).getMethod();
		AccessKind accessKind = target.getAccessKind(m);
		target.getSharedLock(accessKind);
		Object returned = proceed(target);
		target.releaseSharedLock(accessKind);
		return returned;
	}
}