package ca.mcgill.sel.aspectoptimaomtt.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertNotSame;

import org.junit.Test;

import ca.mcgill.sel.aspectoptima.aspects.context.Context;
import ca.mcgill.sel.aspectoptimaomtt.interfacing.AspectOPTIMAOMTT;
import ca.mcgill.sel.aspectoptimaomtt.interfacing.OMTTThread;

public class OMTTTest {
	static final long timeToSleep = 2000;
	static final long timeToWait = timeToSleep + 1000;
	
	@Test
	public void test() { 
		try {
			EnterRunnable r = new EnterRunnable();
			r.triedToEnter = System.currentTimeMillis();
			OMTTThread t = AspectOPTIMAOMTT.beginAndRunTransactionInNewThread(r, "CHECKPOINTING", "TWOPHASELOCKING", true, 2, Integer.MAX_VALUE);
			
			Thread.sleep(timeToWait);
			
			EnterRunnable r2 = new EnterRunnable();
			r2.triedToEnter = System.currentTimeMillis();
			AspectOPTIMAOMTT.spawnRunAndForgetTransactionParticipantFromExistingThread(r2, t);

			Thread.sleep(timeToWait);
			
			long triedToEnter = System.currentTimeMillis();
			AspectOPTIMAOMTT.joinTransactionWithCurrentThread(t);
			long entered = System.currentTimeMillis();
			
			t.join();
			Object c = AspectOPTIMAOMTTTesting.getActualContext();
			System.out.println("c " + c);
			assertNotSame(null, c);
			System.out.println("r.newContext " + r.newContext);
			assertEquals(c, r.newContext);
			System.out.println("AspectOPTIMAOMTTTesting.getActualCollaborativeParticipant() " + AspectOPTIMAOMTTTesting.getActualCollaborativeParticipant());
			System.out.println("r.newPartcipant " + r.newPartcipant);
			assertNotSame(AspectOPTIMAOMTTTesting.getActualCollaborativeParticipant(), r.newPartcipant);
			
			long waiting = r.entered - r.triedToEnter;
			System.out.println("waiting = r.entered - r.triedToEnter " + waiting + " = " + r.entered + " - " + r.triedToEnter);
			assertEquals(true, waiting > timeToWait - timeToSleep);
			
			long waiting2 = r2.entered - r2.triedToEnter;
			System.out.println("waiting2 = r2.entered - r2.triedToEnter " + waiting2 + " = " + r2.entered + " - " + r2.triedToEnter);
			assertEquals(true, waiting2 > timeToWait - timeToSleep);
			
			long waiting3 = entered - triedToEnter;
			System.out.println("waiting3 = entered - triedToEnter " + waiting3 + " = " + entered + " - " + triedToEnter);
			assertEquals(true, waiting3 < timeToWait - timeToSleep);
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
	}
}

class EnterRunnable implements Runnable {
	Object newContext;
	Object newPartcipant;
	long triedToEnter;
	long entered;

	@Override
	public void run() {
		entered = System.currentTimeMillis();
		newContext = AspectOPTIMAOMTTTesting.getActualContext();
		newPartcipant = AspectOPTIMAOMTTTesting.getActualParticipant();
	}
}

class EnterAndSpawnRunnable implements Runnable {
	Object context;
	Object newContext;
	Object newPartcipant;
	long triedToEnter;
	long entered;
	long spawnedEntered;
	
	public EnterAndSpawnRunnable(Object context) {
		this.context = context;
	}

	@Override
	public void run() {
		triedToEnter = System.currentTimeMillis();
		AspectOPTIMAOMTTTesting.joinOMTT(context);
		entered = System.currentTimeMillis();
		newContext = AspectOPTIMAOMTTTesting.getActualContext();
		newPartcipant = AspectOPTIMAOMTTTesting.getActualParticipant();
		System.out.println("newContext " + newContext);
		System.out.flush();
		Runnable r = new Runnable() { public void run() {
			spawnedEntered = System.currentTimeMillis();
			}
		};
		AspectOPTIMAOMTT.spawnAndRunTransactionParticipantFromCurrentThread(r);
	}
}